//	Vamos Automotive Simulator
//  Copyright (C) 2001--2005 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "Three_Vector.h"
#include "Three_Matrix.h"
#include "Two_Vector.h"

using namespace Vamos_Geometry;

const Vamos_Geometry::Three_Vector Vamos_Geometry::
Three_Vector::X = Three_Vector (1.0, 0.0, 0.0);
const Vamos_Geometry::Three_Vector Vamos_Geometry::
Three_Vector::Y = Three_Vector (0.0, 1.0, 0.0);
const Vamos_Geometry::Three_Vector Vamos_Geometry::
Three_Vector::Z = Three_Vector (0.0, 0.0, 1.0);

// Constructors.

// Default constructor.  Null vector.
Vamos_Geometry::
Three_Vector::Three_Vector ()
  : x (0.0),
    y (0.0),
    z (0.0)
{
}

// Elements are initialized by length and angle in the x-y plane.
Vamos_Geometry::
Three_Vector::Three_Vector (double length, double angle)
{
  *this = Three_Vector (length, 0.0, 0.0).rotate (0.0, 0.0, angle);
}

// Elements are initialized by arguments.
Vamos_Geometry::
Three_Vector::Three_Vector (double x_in, double y_in, double z_in)
  : x (x_in),
    y (y_in),
    z (z_in)
{
}

// Initialize with Two_Vector
Vamos_Geometry::
Three_Vector::Three_Vector (const Two_Vector& xy_vector)
  : x (xy_vector.x),
    y (xy_vector.y),
    z (0.0)
{
}

// Zero all elements.
void Vamos_Geometry::
Three_Vector::zero ()
{
  x = y = z = 0.0;
}

// True if all elements are zero.
bool Vamos_Geometry::
Three_Vector::null () const
{
  return (x == 0.0) && (y == 0.0) && (z == 0.0);
}

// Return the dot product with the argument.
double Vamos_Geometry::
Three_Vector::dot (const Three_Vector& vec) const
{
  return x*vec.x + y*vec.y + z*vec.z;
}

// Return the cross product with the argument.
Vamos_Geometry::Three_Vector Vamos_Geometry::
Three_Vector::cross (const Three_Vector& vec) const
{
  return Three_Vector (y * vec.z - z * vec.y,
					   z * vec.x - x * vec.z,
					   x * vec.y - y * vec.x);
}

// Return the projection along the argument.
Vamos_Geometry::Three_Vector Vamos_Geometry::
Three_Vector::project (const Three_Vector& vec) const
{
  double dot_prod = dot (vec);
  double vec_abs = vec.magnitude ();
  if (vec.magnitude () == 0.0)
	return Three_Vector (0.0, 0.0, 0.0);
  return vec.unit () * dot_prod / vec_abs;
}

// Return the vector in the direction of the argument that would yield
// this vector when projected.  This is the inverse of project().
Vamos_Geometry::Three_Vector Vamos_Geometry::
Three_Vector::back_project (const Three_Vector& vec) const
{
  double dot_prod = dot (vec);
  if (dot_prod == 0.0)
	return Three_Vector (0.0, 0.0, 0.0);
  double this_abs = magnitude ();
  return this_abs * this_abs * vec / dot_prod;
}

double Vamos_Geometry::
Three_Vector::perp_distance (const Three_Vector& point1,
							 const Three_Vector& point2) const
{
  Three_Vector vec1 = (point2 - point1);
  Three_Vector vec2 = vec1.project (*this);
  return (vec1 - vec2).magnitude ();
}

double Vamos_Geometry::
Three_Vector::component (const Three_Vector& vec) const
{
  double dot_prod = dot (vec);
  if (dot_prod == 0.0)
	return 0.0;
  return dot_prod / vec.magnitude ();
}  

// Return the unit vector that points along this vector.
Vamos_Geometry::Three_Vector Vamos_Geometry::
Three_Vector::unit () const
{
  Three_Vector vec = *this;
  double vec_abs = magnitude ();
  if (vec_abs == 0.0)
	return Three_Vector (0.0, 0.0, 1.0);
  return vec / vec_abs;
}

// Return the angle between this vector and `vec'.
double Vamos_Geometry::
Three_Vector::angle (const Three_Vector& vec) const
{
  return acos ((*this).dot (vec) / (*this).magnitude () / vec.magnitude ());
}

Vamos_Geometry::Three_Vector& Vamos_Geometry::
Three_Vector::rotate (double x, double y, double z)
{
  Three_Matrix r;
  r.rotate (Three_Vector (x, y, z));
  *this = r * *this; 
  return *this;
}

Vamos_Geometry::Three_Vector Vamos_Geometry::
Three_Vector::rotate (double x, double y, double z) const
{
  Three_Matrix r;
  r.rotate (Three_Vector (x, y, z));
  return r * *this;
}

//* Member operators.
Vamos_Geometry::Three_Vector& Vamos_Geometry::
Three_Vector::operator += (const Three_Vector& vec)
{
  x += vec.x;
  y += vec.y;
  z += vec.z;
  return *this;
}

Vamos_Geometry::Three_Vector& Vamos_Geometry::
Three_Vector::operator -= (const Three_Vector& vec)
{
  x -= vec.x;
  y -= vec.y;
  z -= vec.z;
  return *this;
}

Vamos_Geometry::Three_Vector& Vamos_Geometry::
Three_Vector::operator *= (double factor)
{
  x *= factor;
  y *= factor;
  z *= factor;
  return *this;
}

Vamos_Geometry::Three_Vector& Vamos_Geometry::
Three_Vector::operator /= (double factor)
{
  x /= factor;
  y /= factor;
  z /= factor;
  return *this;
}

//* Non-Member Operators

// Addition
Vamos_Geometry::Three_Vector Vamos_Geometry::
operator + (const Three_Vector& vec1, const Three_Vector& vec2)
{
  Three_Vector out_vec = vec1;
  return out_vec += vec2;
}

// Negation
Vamos_Geometry::Three_Vector Vamos_Geometry::
operator - (const Three_Vector& vec)
{
  Three_Vector out_vec = vec;
  return out_vec *= -1.0;
}

// Subtraction
Vamos_Geometry::Three_Vector Vamos_Geometry::
operator - (const Three_Vector& vec1, const Three_Vector& vec2)
{
  Three_Vector out_vec = vec1;
  return out_vec -= vec2;
}

// Scalar Multiplication
Vamos_Geometry::Three_Vector Vamos_Geometry::
operator * (const Three_Vector& vec, double factor)
{
  Three_Vector out_vec = vec;
  return out_vec *= factor;
}

Vamos_Geometry::Three_Vector Vamos_Geometry::
operator * (double factor, const Three_Vector& vec)
{
  return vec * factor;
}

// Scalar Division
Vamos_Geometry::Three_Vector Vamos_Geometry::
operator / (const Three_Vector& vec, double factor)
{
  Three_Vector out_vec = vec;
  if (factor == 0.0)
	return Three_Vector (0.0, 0.0, 0.0);
  return out_vec /= factor;
}

bool Vamos_Geometry::
operator == (const Three_Vector& vec1, const Three_Vector& vec2)
{
  return (vec1.x == vec2.x)
    && (vec1.y == vec2.y)
    && (vec1.z == vec2.z);
}

// Stream Operators
std::ostream& Vamos_Geometry::
operator << (std::ostream& os, const Three_Vector& vec)
{
  os << "[ " << vec.x 
	 << ", " << vec.y
	 << ", " << vec.z << " ]";
  return os;
}

std::istream& Vamos_Geometry::
operator >> (std::istream& is, Three_Vector& vec)
{
  char delim;
  is >> delim >> vec.x >> delim >> vec.y >> delim >> vec.z >> delim;
  return is;
}
