//	Vamos Automotive Simulator
//  Copyright (C) 2001--2003 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "../geometry/Numeric.h"
#include "../media/Texture_Image.h"
#include "Road_Segment.h"

#include <cmath>
#include <cassert>
#include <iostream>

using namespace Vamos_Geometry;
using namespace Vamos_Track;

//=============================================================================
Kerb::Kerb (const std::vector <Two_Vector>& profile,
			double start, 
			double start_transition_length, 
			double start_transition_width,
			double end, 
			double end_transition_length, 
			double end_transition_width) :
  m_points (profile),
  m_profile (profile),
  m_start (start),
  m_start_transition_length (start_transition_length),
  m_start_transition_width (start_transition_width),
  m_end (end),
  m_end_transition_length (end_transition_length),
  m_end_transition_width (end_transition_width),
  m_full_length (m_end < 0.0)
{
}

void 
Kerb::set_length (double length)
{
  if (m_full_length)
	{
	  m_end = length;
	}
}

double 
Kerb::width () const
{
  return (m_points.size () > 0) ? (m_points.end () - 1)->x : 0.0;
}

bool 
Kerb::on_kerb (double dist) const
{
  return bool ((dist >= m_start) 
			   && ((dist < m_end) || (m_end < 0.0)));
}

double 
Kerb::elevation (double along, double from_inside)
{
  if (!on_kerb (along) || (from_inside < 0.0) || (from_inside > width ()))
	{
	  return 0.0;
	}
  return m_profile.interpolate (from_inside);
}

double 
Kerb::angle (double along, double from_inside)
{
  if (!on_kerb (along) || (from_inside < 0.0) || (from_inside > width ()))
	{
	  return 0.0;
	}
  Two_Vector norm = m_profile.normal (from_inside);
  return std::atan2 (norm.y, norm.x) - pi / 2.0;
}

const Two_Vector& 
Kerb::point (size_t substrip) const
{
  assert (substrip < m_points.size ());
  return m_points [substrip];
}

//=============================================================================
Banking::Banking ()
  : m_start_angle (0.0), m_end_angle (0.0), m_pivot_from_center (0.0),
    mp_bank_curve (new Vamos_Geometry::Spline)
{
}

Banking::~Banking ()
{
  delete mp_bank_curve;
}

void 
Banking::set (const TPoints& angles, double end_angle, double pivot_from_center)
{
  m_angles = angles;
  m_end_angle = end_angle;
  m_pivot_from_center = pivot_from_center;
}

double
Banking::angle (double along) const 
{ 
  return Vamos_Geometry::deg_to_rad (mp_bank_curve->interpolate (along));
}

double 
Banking::height (double along, double from_center) const
{ 
  return (m_pivot_from_center - from_center) * sin (angle (along)); 
}

void 
Banking::set_start (double start_angle, double length)
{
  m_start_angle = start_angle;
  m_angles.clear ();
  m_angles.insert (m_angles.begin (), Two_Vector (0.0, m_start_angle));
  m_angles.push_back (Two_Vector (length, m_end_angle));
  
  delete mp_bank_curve;
  mp_bank_curve = new Vamos_Geometry::Spline (m_angles, 0.0, 0.0);
}

//=============================================================================
Pit_Lane_Transition::Pit_Lane_Transition ()
  : m_direction (IN),
    m_side (LEFT),
    m_split_or_join (0.0),
    m_merge (0.0),
    m_pit_width (0.0),
    m_pit_shoulder_width (0.0),
    m_merge_is_set (false),
    m_width_is_set (false)
{
}

void
Pit_Lane_Transition::set_merge (Direction direction,
                                Direction side,
                                double split_or_join,
                                double merge)
{
  m_direction = direction;
  m_side = side;
  m_split_or_join = split_or_join;
  m_merge = merge;
  m_merge_is_set = true;
}

void
Pit_Lane_Transition::set_width (double pit_width, 
                                double left_shoulder,
                                double right_shoulder)
{
  m_pit_width = pit_width;
  m_pit_shoulder_width = (m_side == LEFT) ? left_shoulder : right_shoulder;
  m_width_is_set = true;
}

double
Pit_Lane_Transition::width (Direction pit_side, double distance, bool narrow) const
{
  if (pit_side != m_side) return 0.0;
  if (narrow) return m_pit_width;
  if (((m_direction == IN) && distance <= m_split_or_join)
      || ((m_direction == OUT) && distance >= m_split_or_join))
    return 0.0;
  return m_pit_width;
}

void
Pit_Lane_Transition::scale (double factor)
{
  m_split_or_join *= factor;
  m_merge *= factor;
}

//=============================================================================
Road_Segment::Road_Segment (double length,
                            double radius,
                            double left_width,
                            double right_width,
                            double left_road_width,
                            double right_road_width)
  : m_length (length),
    m_radius (radius),
    m_left_wall_height (0.0),
    m_right_wall_height (0.0),
    m_start_skew (0.0),
    m_end_skew (0.0),
    mp_elevation_curve (0),
    mp_left_kerb (0),
    mp_right_kerb (0),
    m_start_angle (0.0),
    m_last_segment (false)
{
  m_left_road_width.load (Two_Vector (0.0, left_road_width));
  m_right_road_width.load (Two_Vector (0.0, right_road_width));
  m_left_width.load (Two_Vector (0.0, left_width));
  m_right_width.load (Two_Vector (0.0, right_width));
}

Road_Segment::~Road_Segment ()
{
  delete mp_left_kerb;
  delete mp_right_kerb;
}

void
Road_Segment::set_widths (const TPoints& right, const TPoints& right_road,
                          const TPoints& left_road, const TPoints& left)
{
  m_right_width.replace (right);
  m_right_road_width.replace (right_road);
  m_left_road_width.replace (left_road);
  m_left_width.replace (left);
}

void
Road_Segment::set_kerb (Kerb* kerb, Direction side)
{
  if (side == LEFT)
    {
      delete mp_left_kerb;
      mp_left_kerb = kerb;
    }
  else
    {
      delete mp_right_kerb;
      mp_right_kerb = kerb;
    }
}

double
Road_Segment::kerb_width (Vamos_Geometry::Direction side, double along) const
{
  const Kerb* kerb = (side == LEFT) ? mp_left_kerb : mp_right_kerb;
  if ((kerb != 0) && kerb->on_kerb (along))
    return kerb->width ();
  return 0.0;
}

void
Road_Segment::set_wall_heights (double left_height, double right_height)
{
  m_left_wall_height = left_height;
  m_right_wall_height = right_height;
}

void 
Road_Segment::build_elevation (Spline* elevation,
							   double start_distance)
{
  mp_elevation_curve = elevation;
  for (std::vector <Two_Vector>::iterator it = m_elevation_points.begin ();
	   it != m_elevation_points.end ();
	   it++)
	{
	  mp_elevation_curve->load (*it + Two_Vector (start_distance, 0.0));
	}

  if (m_last_segment)
	{
	  // Remove any elevation points near or greater than m_length.
	  mp_elevation_curve->remove_greater (start_distance + m_length - 10.0);
	  // Force the elevation curve to return to zero.
	  mp_elevation_curve->load (Two_Vector (start_distance + m_length, 0.0)); 
	}
}

double 
Road_Segment::elevation (double along, double from_center) const
{
  assert (mp_elevation_curve != 0);
  double elev = mp_elevation_curve->interpolate (along + m_start_distance)
	+ banking().height (along, from_center);

  // Add the kerb's height.
  double diff = from_center - left_road_width (along);
  if (mp_left_kerb)
    elev += mp_left_kerb->elevation (along, diff);
  diff = -from_center - right_road_width (along);
  if (mp_right_kerb)
    elev += mp_right_kerb->elevation (along, diff);

  return elev;
}

Three_Vector 
Road_Segment::barrier_normal (double along, double from_center) const
{
  return ((from_center > 0.0) 
          ? -m_left_width.normal (along)
          : m_right_width.normal (along)).
    rotate (0.0, 0.0, angle (along));
}

Three_Vector 
Road_Segment::normal (double along, double from_center) const
{
  assert (mp_elevation_curve != 0);

  Three_Vector norm (mp_elevation_curve->normal (along + m_start_distance));
  norm.z = norm.y;
  norm.y = 0.0;

  double bank = m_banking.angle (along);
  if (mp_left_kerb)
    bank -= mp_left_kerb->angle (along, from_center - left_road_width (along));
  if (mp_right_kerb)
    bank += mp_right_kerb->angle (along, -from_center - right_road_width (along));
  norm.rotate (-bank, 0.0, 0.0);
  return norm.rotate (0.0, 0.0, angle (along));
}

void 
Road_Segment::set_start (const Three_Vector& start_coords, 
                         double start_distance,
                         double start_angle,
                         double start_bank,
                         const std::vector <double>& texture_offsets)
{
  m_start_distance = start_distance;
  m_start_coords = start_coords;
  m_start_angle = branch (start_angle, 0.0);
  m_banking.set_start (start_bank, m_length);
}

Three_Vector
Road_Segment::center_of_curve () const
{
  return m_start_coords + Three_Vector (m_radius, m_start_angle + pi/2);
}

Three_Vector
Road_Segment::end_coords () const
{
  if (is_straight ())
    return m_start_coords + Three_Vector (m_length, m_start_angle);
  else
    return center_of_curve () 
      - Three_Vector (m_radius, m_start_angle + arc () + pi/2);
}

double 
Road_Segment::left_width (double distance, bool narrow) const
{
  return m_left_width.interpolate (distance)
    - m_pit.width (LEFT, distance, narrow);
}

double 
Road_Segment::right_width (double distance, bool narrow) const
{
  return m_right_width.interpolate (distance)
    - m_pit.width (RIGHT, distance, narrow);
}

void 
Road_Segment::set_left_width (double distance, double width)
{
  m_left_width.load (Two_Vector (distance, width));
}

void 
Road_Segment::set_right_width (double distance, double width)
{
  m_right_width.load (Two_Vector (distance, width));
}

void 
Road_Segment::set_left_road_width (double distance, double width)
{
  m_left_road_width.load (Two_Vector (distance, width));
}

void 
Road_Segment::set_right_road_width (double distance, double width)
{
  m_right_road_width.load (Two_Vector (distance, width));
}

double
Road_Segment::pit_width () const
{
  Direction side = m_pit.side ();
  double distance = m_pit.split_or_join ();
  double width = (side == LEFT)
    ? m_left_width.interpolate (distance)
    : m_right_width.interpolate (distance);
  return width - m_pit.width (side, distance, true);
}

double
Road_Segment::pit_road_connection () const
{
  return m_pit.split_or_join ();
}

double
Road_Segment::extra_road_width (Direction pit_side, 
                                double distance,
                                bool narrow) const
{
  if (narrow || (pit_side != m_pit.side ())) return 0.0;

  double width = -m_pit.shoulder_width ();
  if (pit_side == LEFT)
    width += left_width (distance) - left_road_width (distance, true);
  else
    width += right_width (distance) - right_road_width (distance, true);

  const double extra = width * (distance - m_pit.merge ())
    / (m_pit.split_or_join () - m_pit.merge ());
   
  if ((m_pit.direction () == IN)
      && (distance > m_pit.merge ())
      && (distance <= m_pit.split_or_join ()))
    return extra;
  if ((m_pit.direction () == OUT)
      && (distance < m_pit.merge ())
      && (distance >= m_pit.split_or_join ()))
    return extra;

  return 0.0;
}

double 
Road_Segment::left_road_width (double distance, bool narrow) const
{
  return m_left_road_width.interpolate (distance)
    + extra_road_width (LEFT, distance, narrow);
}

double 
Road_Segment::right_road_width (double distance, bool narrow) const
{
  return m_right_road_width.interpolate (distance)
    + extra_road_width (RIGHT, distance, narrow);
}

const 
Pit_Lane_Transition& Road_Segment::pit () const
{
  return m_pit;
}

void 
Road_Segment::set_pit_width (double width, 
                             double left_shoulder, 
                             double right_shoulder)
{
  m_pit.set_width (width, left_shoulder, right_shoulder);
}

void 
Road_Segment::set_pit_lane (Direction direction,
                            Direction side, 
                            double split_or_join, 
                            double merge)
{
  m_pit.set_merge (direction, side, split_or_join, merge);
}

void 
Road_Segment::set_length (double length)
{
  scale_widths (length / m_length);
  m_length = length;
}

void 
Road_Segment::scale (double factor)
{
  assert (factor != 0);
  scale_widths (factor);
  m_length *= factor;
  m_radius *= factor;
  m_pit.scale (factor);
}

void
Road_Segment::scale_widths (double factor)
{
  m_left_width.scale (factor);
  m_right_width.scale (factor);
  m_left_road_width.scale (factor);
  m_right_road_width.scale (factor);
}

std::complex <double> 
solve_quadratic (double a, double b, double c, double solution)
{
  if (a == 0.0) return -c / b;
  return (-b + Vamos_Geometry::sign (solution) 
          * sqrt (std::complex <double> (b*b - 4*a*c))) 
    / (2*a);
}

// Find the angle from the beginning of a curve.
double 
Road_Segment::get_curve_angle (const Three_Vector& position,
                               double across)
{
  return arc () / 2.0 
    + atan2 (sign (m_radius) * position.y,
             sign (m_radius) 
             * (position.x - across * m_start_skew / sin (arc () / 2.0)));
}

// Do the world-to-track coordinate transformation.
double
Road_Segment::coordinates (const Three_Vector& world_position,
						   Three_Vector& track_position)
{
  if (!is_straight ())
    {
      const double half_angle = arc () / 2.0;

      const Three_Vector centered_position = 
        (world_position - center_of_curve ()).
        rotate (0.0, 0.0, pi / 2.0 - half_angle - m_start_angle);

      const std::complex <double> across =
        solve_quadratic (1.0 + 
                         2.0 * m_start_skew / tan (half_angle)
                         - m_start_skew * m_start_skew,
                         -2.0 * (m_radius + (m_start_skew / sin (half_angle)
                                             * (m_radius * cos (half_angle)
                                                - centered_position.x))),
                         m_radius * m_radius
                         - centered_position.x * centered_position.x
                         - centered_position.y * centered_position.y,
                         -m_radius);

      track_position.y = across.real ();
      if (across.imag () != 0.0)
        {
          // We're off the segment but the solution is complex.  Just
          // indicate whether we're off the beginning or end of the
          // segement.
          if (centered_position.y * m_radius < 0.0)
            track_position.x = m_length + 1.0;
          else
            track_position.x = -1.0;
        }
      else
        {
          track_position.x = m_radius
            * get_curve_angle (centered_position, track_position.y);
        }
      track_position.z = 0.0;
    }
  else
    {
      track_position = 
        (world_position - center_of_curve ()).rotate (0.0, 0.0, -m_start_angle);
      track_position.x = (track_position.x - track_position.y * m_start_skew)/
        (1 + track_position.y / m_length * (m_end_skew - m_start_skew));
    }
  track_position.z = elevation (track_position.x, track_position.y);
  return off_track_distance (track_position);
} 

// How far is the position off the end (+ve) or beginning (-ve) of
// the segment?  Return 0 if the position is on the segment.
double
Road_Segment::off_track_distance (const Three_Vector& track_position) const
{
  const double max
    = ((m_pit.direction () == IN)
       && on_pit_merge (track_position.x, track_position.y))
    ? m_pit.split_or_join () : m_length;

  const double min
    = ((m_pit.direction () == OUT) 
       && on_pit_merge (track_position.x, track_position.y))
    ? m_pit.split_or_join () : 0.0;

  if (track_position.x < min)
    return track_position.x - min;
  if (track_position.x > max)
    return track_position.x - max;
  return 0.0;
}

bool
Road_Segment::on_pit_merge (double distance, double from_center) const
{
#warning "determine if the car is on the pit-lane merge."
  return false;
}

// Do the track-to-world coordinate transformation.
Three_Vector
Road_Segment::position (double along, double from_center) const
{
  Three_Vector pos (0.0, 0.0, elevation (along, from_center));
  if (is_straight ())
    {
      pos += start_coords () 
        + Three_Vector (along, m_start_angle)
        + Three_Vector (from_center, m_start_angle + pi/2);
    }
  else
    {
      const double angle = m_start_angle + along / m_radius;
      pos += center_of_curve () 
        + Three_Vector (m_radius - from_center, angle - pi/2);
    }
  return pos;
}
