//  Strip_Track.cc - a track.
//
//	Vamos Automotive Simulator
//  Copyright (C) 2001--2004 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "Strip_Track.h"
#include "../geometry/Numeric.h"
#include "../geometry/Spline.h"
#include "../media/Texture_Image.h"

#include <GL/glu.h>

#include <cmath>
#include <cassert>

using namespace Vamos_Geometry;
using namespace Vamos_Media;
using namespace Vamos_Track;

Camera s_default_camera (0, Three_Vector (100.0, -20.0, 10.0), 0.0);

namespace Vamos_Track
{
  class Sky_Box
  {
    Texture_Image* mp_sides;
    Texture_Image* mp_top;
    Texture_Image* mp_bottom;
	GLuint m_list_id;

  public:
    Sky_Box (double side_length,
             std::string sides_image,
             std::string top_image,
             std::string bottom_image,
             bool smooth);
    ~Sky_Box ();

    void draw (const Three_Vector& view) const;
  };
}

Sky_Box::Sky_Box (double side_length,
                  std::string sides_image,
                  std::string top_image,
                  std::string bottom_image,
                  bool smooth)
  // Clamp the textures to aviod showing seams.
  : mp_sides (new Texture_Image (sides_image, smooth, true, 
                                 GL_CLAMP_TO_EDGE)),
    mp_top (new Texture_Image (top_image, smooth, true, 
                               GL_CLAMP_TO_EDGE)),
    mp_bottom (new Texture_Image (bottom_image, smooth, true, 
                                  GL_CLAMP_TO_EDGE)),
    m_list_id (glGenLists (1))
{
  double height = side_length;
  double length = side_length;
  double width = side_length;

  double x = -length / 2.0;
  double y = -width / 2.0;
  double z = -height / 2.0;

  glNewList (m_list_id, GL_COMPILE);
  glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  mp_sides->activate ();

  // front
  glBegin (GL_QUAD_STRIP);
  glTexCoord2d (0.0, 0.0); 
  glVertex3d (x + length, y + width, z + height);
  glTexCoord2d (0.0, 1.0); 
  glVertex3d (x + length, y + width, z);
  glTexCoord2d (0.25, 0.0); 
  glVertex3d (x + length, y, z + height); 
  glTexCoord2d (0.25, 1.0); 
  glVertex3d (x + length, y, z);

  // right
  glTexCoord2d (0.25, 0.0); 
  glVertex3d (x + length, y, z + height);
  glTexCoord2d (0.25, 1.0); 
  glVertex3d (x + length, y, z);
  glTexCoord2d (0.5, 0.0); 
  glVertex3d (x, y, z + height); 
  glTexCoord2d (0.5, 1.0); 
  glVertex3d (x, y, z);

  // back
  glTexCoord2d (0.5, 0.0); 
  glVertex3d (x, y, z + height);
  glTexCoord2d (0.5, 1.0); 
  glVertex3d (x, y, z);
  glTexCoord2d (0.75, 0.0); 
  glVertex3d (x, y + width, z + height); 
  glTexCoord2d (0.75, 1.0); 
  glVertex3d (x, y + width, z);

  // left
  glTexCoord2d (0.75, 0.0); 
  glVertex3d (x, y + width, z + height);
  glTexCoord2d (0.75, 1.0); 
  glVertex3d (x, y + width, z);
  glTexCoord2d (1.0, 0.0); 
  glVertex3d (x + length, y + width, z + height); 
  glTexCoord2d (1.0, 1.0); 
  glVertex3d (x + length, y + width, z);
  glEnd();

  // top
  mp_top->activate ();

  glBegin (GL_QUADS);		
  glTexCoord2d (0.0, 0.0); 
  glVertex3d (x, y + width, z + height);
  glTexCoord2d (0.0, 1.0); 
  glVertex3d (x + length, y + width, z + height); 
  glTexCoord2d (1.0, 1.0); 
  glVertex3d (x + length, y, z + height);
  glTexCoord2d (1.0, 0.0); 
  glVertex3d (x, y, z + height);
  glEnd ();

  // bottom
  mp_bottom->activate ();

  glBegin (GL_QUADS);		
  glTexCoord2d (0.0, 0.0); 
  glVertex3d (x + length, y + width, z);
  glTexCoord2d (0.0, 1.0); 
  glVertex3d (x, y + width, z);
  glTexCoord2d (1.0, 1.0); 
  glVertex3d (x, y, z);
  glTexCoord2d (1.0, 0.0); 
  glVertex3d (x + length, y, z); 
  glEnd ();

  glFlush ();
  glEndList ();
}

Sky_Box::~Sky_Box ()
{
  delete mp_bottom;
  delete mp_top;
  delete mp_sides;
  glDeleteLists (m_list_id, 1);
}

void
Sky_Box::draw (const Three_Vector& view) const
{
  glLoadIdentity ();
  glTranslatef (view.x, view.y, view.z);
  glCallList (m_list_id);
  // Clear the depth buffer to keep the sky behind everything else.
  glClear (GL_DEPTH_BUFFER_BIT);
}

//* Class Map_Background

namespace Vamos_Track
{
  class Map_Background
  {
    Texture_Image* mp_image;
    double m_x_offset;
    double m_y_offset;
    double m_x_size;
    double m_y_size;

  public:
    Map_Background (std::string image_file_name,
                    double x_offset, double y_offset,
                    double x_size, double y_size);
    ~Map_Background ();

    void draw () const;
  };
}

Map_Background::Map_Background (std::string image_file_name,
                                double x_offset, double y_offset,
                                double x_size, double y_size)
  : mp_image (new Texture_Image (image_file_name, true)),
    m_x_offset (x_offset),
    m_y_offset (y_offset),
    m_x_size (x_size),
    m_y_size (y_size)
{
}

Map_Background::~Map_Background ()
{
  delete mp_image;
}

void
Map_Background::draw () const
{
  glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  mp_image->activate ();

  glLoadIdentity ();
  glTranslatef (m_x_offset, m_y_offset, 0.0);

  glBegin (GL_QUADS);
  glTexCoord2d (0.0, 1.0);
  glVertex3d (m_x_offset, m_y_offset, 0.0);
  glTexCoord2d (0.0, 0.0); 
  glVertex3d (m_x_offset, m_y_offset + m_y_size, 0.0);
  glTexCoord2d (1.0, 0.0); 
  glVertex3d (m_x_offset + m_x_size, m_y_offset + m_y_size, 0.0); 
  glTexCoord2d (1.0, 1.0); 
  glVertex3d (m_x_offset + m_x_size, m_y_offset, 0.0);
  glEnd ();

  // Clear the depth buffer to keep the background behind the track.
  glClear (GL_DEPTH_BUFFER_BIT);
}

//* Struct Camera
Camera::Camera (size_t segment_index_in,
                const Three_Vector& position_in,
                double range_in)
  : segment_index (segment_index_in),
    position (position_in),
    fixed (false),
    vertical_field_angle (10.0),
    range (range_in)
{}

Camera::Camera () 
  : segment_index (0), 
    fixed (false),
    vertical_field_angle (10.0), 
    range (0.0) 
{}

//* Class Road
Road::Road ()
  : mp_elevation (new Spline (0.0, 0.0)),
    m_start_direction (0.0)
{
  clear ();
}

Road::~Road ()
{
  clear ();
  delete mp_elevation;
}

void
Road::clear ()
{
  mp_elevation->clear ();
  mp_elevation->load (Two_Vector (0.0, 0.0));
  m_length = 0.0;
  m_bounds = Rectangle (0.0, 0.0, 0.0, 0.0);

  for (Segment_List::iterator it = m_segments.begin ();
	   it != m_segments.end ();
	   it++)
	{
	  delete (*it);
	}
  m_segments.clear ();
}

size_t
Road::add_segment (Gl_Road_Segment* segment)
{
  m_segments.push_back (segment);
  return m_segments.size ();
}

double 
Road::build_elevation ()
{
  double length = 0.0;
  for (Segment_List::iterator it = m_segments.begin ();
	   it != m_segments.end ();
	   it++)
	{
	  (*it)->build_elevation (mp_elevation, length);
	  length += (*it)->length ();
	}
  return length;
}

void
Road::build_segments (Three_Vector start_coords, 
                      double start_angle, 
                      double start_bank)
{
  std::vector <double> 
    texture_offsets ((*(m_segments.begin ()))->materials ().size ());

  narrow_pit_segments ();

  m_length = 0.0;
  for (Segment_List::iterator it = m_segments.begin ();
	   it != m_segments.end ();
	   it++)
	{
      (*it)->set_start (start_coords, m_length, start_angle, start_bank,
                        texture_offsets);
	  (*it)->build ();

	  // Update the bounding dimensions.
      m_bounds.enclose ((*it)->bounds ());
      m_length += (*it)->length ();

	  start_coords = (*it)->end_coords ();
	  start_angle = (*it)->end_angle ();
	  start_bank = (*it)->banking ().end_angle ();
	  texture_offsets = (*it)->texture_offsets ();
	}
}

void
Road::narrow_pit_segments ()
{
  for (Segment_List::iterator it = m_segments.begin ();
	   it != m_segments.end ();
	   it++)
	{
      const Pit_Lane_Transition& pit = (*it)->pit ();
      if (!pit.active ())
        continue;
      if (pit.direction () == OUT)
        {
          for (Segment_List::reverse_iterator rit (it);
               ((rit != m_segments.rend ())
                  && !(*rit)->pit ().active ());
               rit++)
            {
              // rit (it) initializes rit to point to the element
              // before *it.
              (*rit)->set_pit_lane (pit.direction (),
                                    pit.side (),
                                    (*rit)->length () + 1.0,
                                    (*rit)->length () + 1.0);
              (*rit)->set_pit_width ((*it)->pit_width (), 0.0, 0.0);
            }
        }
      else
        {
          for (Segment_List::iterator it2 (it + 1);
               ((it2 != m_segments.end ())
                && !(*it2)->pit ().active ());
               it2++)
            {
              (*it2)->set_pit_lane (pit.direction (), pit.side (), -1.0, -1.0);
              (*it2)->set_pit_width ((*it)->pit_width (), 0.0, 0.0);
            }
       }
    }
}

void 
Road::build (bool close, int adjusted_segments, double length)
{
  set_skews ();
  if (close)
    {
      join ((*m_segments.begin ())->start_coords (), 
            (*m_segments.begin ())->start_angle (),
            (*m_segments.begin ())->start_coords (), 
            (*m_segments.begin ())->start_angle (),
            adjusted_segments);
      // Force the segment to end at 0, 0.
      (*(m_segments.end () - 1))->last_segment (true);
    }
  if (length != 0.0)
      set_length (length);

  build_elevation ();
  build_segments (Three_Vector (), start_direction (), 0.0);
}

double perpendicular_distance (const Three_Vector& p1, 
                               const Three_Vector& p2,
                               double angle)
{
  return (p2 - p1).magnitude () 
            * sin (atan2 (p1.y - p2.y, p1.x - p2.x) - angle);
}

// Force a road that starts at START_COORDS and START_ANGLE, to end at
// END_COORDS and END_ANGLE.
void 
Road::join (const Three_Vector& start_coords,
            double start_angle,
            const Three_Vector& end_coords,
            double end_angle,
            int adjusted_segments)
{
  if (adjusted_segments == 0) return;
  if ((adjusted_segments < 0) || (adjusted_segments > 3))
    throw Can_Not_Close ();

  // Find the position and angle at the (unajusted) end of the road.
  Three_Vector end (start_coords.x, start_coords.y, 0.0);
  double angle = start_angle;
  Three_Vector center;
  for (Segment_List::iterator it = m_segments.begin ();
	   it != m_segments.end () - (adjusted_segments > 1 ? 2 : 1);
	   it++)
	{
      if ((*it)->radius () == 0.0)
        {
          end += Three_Vector ((*it)->length (), angle);
        }
      else
        {
          center = end + Three_Vector ((*it)->radius (), angle + pi/2.0);
          angle += (*it)->arc ();
          end = center - Three_Vector ((*it)->radius (), angle + pi/2.0);
        }
    }

  if (adjusted_segments > 1)
    {
      Gl_Road_Segment* last_curve = *(m_segments.end () - 2);
	  if (last_curve->radius () == 0.0)
        throw Can_Not_Close ();

      // Change the length, but not the radius of the last curve to
      // make the last straight and parallel with the beginning of the
      // first segment.
      last_curve->set_arc (branch (end_angle - angle, -pi));

      center = end + Three_Vector (last_curve->radius (), angle + pi/2.0);
      
      if (adjusted_segments == 3)
        {
          Gl_Road_Segment* other_straight = *(m_segments.end () - 3);
          if (other_straight->radius () != 0.0)
            throw Can_Not_Close ();

          double delta_length = 
            (last_curve->radius () 
             - perpendicular_distance (center, end_coords, end_angle)) 
            / sin (angle - end_angle);
          other_straight->set_length (other_straight->length () + delta_length);
          end += Three_Vector (delta_length, angle);
        }
      else if (adjusted_segments == 2)
        {
          last_curve->scale (perpendicular_distance (end, end_coords, end_angle)
                             / (last_curve->radius () 
                                * (1 - cos (last_curve->arc ()))));
        }

      center = end + Three_Vector (last_curve->radius (), angle + pi/2.0);
      angle += last_curve->arc ();
      end = center - Three_Vector (last_curve->radius (), angle + pi/2.0);
    }

  Gl_Road_Segment* last_straight = *(m_segments.end () - 1);
  if (last_straight->radius () != 0.0)
    throw Can_Not_Close ();
  last_straight->set_length ((end_coords - end).magnitude ());
}

void 
Road::set_skews ()
{
  for (Segment_List::iterator it = m_segments.begin () + 1;
	   it != m_segments.end ();
	   it++)
	{
      double skew = (*it)->start_skew ();
      if ((skew != 0.0) && ((*it)->arc () != 0.0))
        {
          if ((*(it - 1))->arc () == 0.0)
            (*(it - 1))->set_end_skew (skew);
          if ((*(it + 1))->arc () == 0.0)
            (*(it + 1))->set_start_skew (-skew);
        }
    }
}

// Scale the track to a particular length.
void 
Road::set_length (double length)
{
  assert (m_segments.size () != 0);

  // Find the current length.
  double old_length = 0.0;
  for (Segment_List::iterator it = m_segments.begin ();
	   it != m_segments.end ();
	   it++)
	{
	  old_length += (*it)->length ();
	}

  assert (old_length != 0.0);
  double factor = length / old_length;

  // Adjust the segments.
  for (Segment_List::iterator it = m_segments.begin ();
	   it != m_segments.end ();
	   it++)
	{
	  (*it)->scale (factor);
	}
}

void
Road::set_start_slope (double slope)
{
  mp_elevation->set_slope (slope, slope);
}

void
Road::draw ()
{
  std::for_each (m_segments.begin (), m_segments.end (), 
                 std::mem_fun (&Gl_Road_Segment::draw));
}

Three_Vector
Pit_Lane::pit_in_offset (Gl_Road_Segment& pit_in) const
{
  double pit_width;
  if (pit_in.pit ().side () == LEFT)
    pit_width = (*segments ().begin ())->left_width (0.0);
  else
    pit_width = (*segments ().begin ())->right_width (0.0);
  pit_width /= cos (start_direction ());

  double along = pit_in.pit ().split_or_join ();
  double offset;
  if (pit_in.pit ().side () == LEFT)
    offset = pit_in.left_width (along) - pit_width;
  else
    offset = -pit_in.right_width (along) + pit_width;

  if (pit_in.radius () == 0.0)
    return Three_Vector (along, offset, 0.0).
      rotate (0.0, 0.0, pit_in.angle (along));
  else
    return pit_in.center_of_curve () - pit_in.start_coords ()
      + Three_Vector (pit_in.radius () - offset, 
                      pit_in.angle (along) - pi/2);
}

Three_Vector
Pit_Lane::pit_out_offset (Gl_Road_Segment& pit_out) const
{
  double pit_width;
  const Gl_Road_Segment& last_segment = **(segments ().end () - 1); 
  if (pit_out.pit ().side () == LEFT)
    pit_width = last_segment.left_width (last_segment.length ());
  else
    pit_width = last_segment.right_width (last_segment.length ());
  pit_width /= cos (start_direction ());

  double along = pit_out.pit ().split_or_join ();
  double offset;
  if (pit_out.pit ().side () == LEFT)
    offset = pit_out.left_width (along) - pit_width;
  else
    offset = -pit_out.right_width (along) + pit_width;

  if (pit_out.radius () == 0.0)
    return Three_Vector (along, offset, 0.0).
      rotate (0.0, 0.0, pit_out.angle (along));
  else
    return pit_out.center_of_curve () - pit_out.start_coords ()
      + Three_Vector (pit_out.radius () - offset, 
                      pit_out.angle (along) - pi/2);
}

void
Pit_Lane::build (bool join_to_track, 
                 int adjusted_segments,
                 Gl_Road_Segment& pit_in,
                 Gl_Road_Segment& pit_out,
                 const Spline& track_elevation)
{
  if (m_segments.size () == 0) return;

  set_skews ();
  double pit_skew = tan (start_direction ());

  (*m_segments.begin ())->set_start_skew (pit_skew);
  (*(m_segments.end () - 1))->set_end_skew (tan (end_direction ()));

  if (join_to_track)
    {
      join (pit_in.start_coords () + pit_in_offset (pit_in), 
            pit_in.pit_angle () + start_direction (),
            pit_out.start_coords () + pit_out_offset (pit_out),
            pit_out.pit_angle () + end_direction (),
            adjusted_segments);
    }

  delete mp_elevation;
  double in_distance = 
    pit_in.start_distance () + pit_in.pit ().split_or_join ();
  double out_distance = 
    pit_out.start_distance () + pit_out.pit ().split_or_join ();
  mp_elevation = new Spline (track_elevation.slope (in_distance),
                             track_elevation.slope (out_distance));
  mp_elevation->
    load (Two_Vector (0.0, track_elevation.interpolate (in_distance)));
  m_length = build_elevation ();
  mp_elevation->
    load (Two_Vector (m_length, track_elevation.interpolate (out_distance)));

  build_segments (pit_in.start_coords () + pit_in_offset (pit_in),
                  pit_in.pit_angle () + start_direction (),
                  pit_in.banking ().end_angle ());
}

//* Class Strip_Track
Strip_Track::Strip_Track () :
  mp_track (new Road),
  mp_pit_lane (new Pit_Lane),
  m_pit_in_index (-1),
  m_pit_out_index (-1),
  mp_sky_box (0),
  mp_map_background (0)
{
  m_timing_lines.clear ();
  m_cameras.clear ();
}


Strip_Track::~Strip_Track ()
{
  delete mp_pit_lane;
  delete mp_track;
  delete mp_sky_box;
  delete mp_map_background;
}

// Read the track definition file.
void 
Strip_Track::read (std::string data_dir, std::string track_file)
{
  // Remember the file name for re-reading.
  if ((data_dir != "") && (track_file != ""))
	{
	  m_data_dir = data_dir;
	  m_track_file = track_file;
	}

  mp_track->clear ();
  mp_pit_lane->clear ();
  m_timing_lines.clear ();

  Strip_Track_Reader reader (m_data_dir, m_track_file, this);
}

// Add a straight or curved segment of road to the track.
size_t
Strip_Track::add_segment (Gl_Road_Segment* segment)
{
  return mp_track->add_segment (segment);
}

size_t
Strip_Track::add_pit_segment (Gl_Road_Segment* segment)
{
  const bool start = (mp_pit_lane->segments ().size () == 0);
  const double distance = start ? 0.0 : segment->length ();
  const double width = 
    segment->left_width (distance) + segment->right_width (distance);
  const double left_shoulder = 
    segment->left_width (distance) - segment->left_road_width (distance);
  const double right_shoulder = 
    segment->right_width (distance) - segment->right_road_width (distance);

  if (start)
    {
      mp_track->segments ()[m_pit_in_index]->
        set_pit_width (width, left_shoulder, right_shoulder);
    }
  else
    {
      mp_track->segments ()[m_pit_out_index]->
        set_pit_width (width, left_shoulder, right_shoulder);
    }
  return mp_pit_lane->add_segment (segment);
}

void
Strip_Track::set_pit_angles (double in, double out)
{
  mp_pit_lane->set_start_direction (in);
  mp_pit_lane->set_end_direction (out);
}

// Make the track.
void 
Strip_Track::build (bool close,
                    int adjusted_road_segments, 
                    double track_length, 
                    bool join_pit_lane,
                    int adjusted_pit_segments)
{
  mp_track->build (close, adjusted_road_segments, track_length);

  if ((m_pit_in_index != -1) && (m_pit_out_index != -1))
    {
      mp_pit_lane->build (join_pit_lane, 
                          adjusted_pit_segments,
                          *mp_track->segments ()[m_pit_in_index],
                          *mp_track->segments ()[m_pit_out_index],
                          mp_track->elevation ());
    }
}

void
Strip_Track::set_sky_box (std::string sides_image, 
                          std::string top_image,
                          std::string bottom_image,
                          bool smooth)
{
  delete mp_sky_box;
  mp_sky_box = new Sky_Box (100.0, 
                            sides_image, 
                            top_image, 
                            bottom_image,
                            smooth);
}

void
Strip_Track::set_map_background (std::string background_image,
                                 double x_offset, double y_offset,
                                 double x_size, double y_size)
{
  delete mp_map_background;
  mp_map_background = new Map_Background (background_image,
                                          x_offset, y_offset,
                                          x_size, y_size);
}

void
Strip_Track::draw_sky (const Three_Vector& view) const
{
  mp_sky_box->draw (view);
}

void
Strip_Track::draw_map_background () const
{
  if (mp_map_background != 0)
    mp_map_background->draw ();
}

void 
Strip_Track::draw () const
{
  glLoadIdentity ();
  mp_track->draw ();
  mp_pit_lane->draw ();
}

// Return the new position for a vehicle at POS when a reset is
// performed.
Three_Vector 
Strip_Track::reset_position (const Three_Vector& pos,
							 size_t& road_index,
							 size_t& segment_index)
{
  const Three_Vector& track_pos = 
    track_coordinates (pos, road_index, segment_index);

  const Gl_Road_Segment* segment = 
    get_road (road_index).segments ()[segment_index];
  const double along = track_pos.x - segment->start_distance ();
  Three_Vector direction = segment->barrier_normal (along, -1.0);
  direction.z = -sin (segment->banking ().angle (along));

  const double across = track_pos.y;
  return pos - across * direction;
}

// Return the new orientation for a vehicle at POS when a reset is
// performed.
Three_Matrix 
Strip_Track::reset_orientation (const Three_Vector& pos,
								size_t& road_index,
								size_t& segment_index)
{
  Three_Matrix orientation;
  orientation.identity ();

  // Align the car's up direction with the normal.
  const Three_Vector& track_pos = 
    track_coordinates (pos, road_index, segment_index);

  const Gl_Road_Segment* segment = 
    get_road (road_index).segments ()[segment_index];
  const double along = track_pos.x - segment->start_distance ();
  const double across = track_pos.y;
  Three_Vector normal = segment->normal (along, across); 
  orientation.rotate (Three_Vector (-asin (normal.y), 
									asin (normal.x),
									segment->angle (along)));
  return orientation;
}

// Return the elevation of the track at the x and y components of POS.
double 
Strip_Track::elevation (const Three_Vector& pos, 
						double bump_parameter,
						size_t& road_index,
						size_t& segment_index)
{
  return track_coordinates (pos, road_index, segment_index).z
	+ m_material.bump (bump_parameter);
}

// Return WORLD_POS transformed to the track's coordinate system.
// SEGMENT_INDEX will be modified if the position on another
// segment.
Three_Vector 
Strip_Track::track_coordinates (const Three_Vector& world_pos,
								size_t& road_index,
								size_t& segment_index)
{
  // Find the distance along the track, distance from center, and elevation
  // for the world coordinates `world_pos.x' and `world_pos.y'.

  Three_Vector track_pos;
  const Segment_List* segments = &get_road (road_index).segments ();
  assert (segment_index < segments->size ());
  Gl_Road_Segment* segment = (*segments)[segment_index];

  size_t i = 0;
  bool found = false;
  while (i < segments->size () + 1)
	{
	  double off = segment->coordinates (world_pos, track_pos); 
	  if (off == 0.0)
		{
          found = true;
		  break;
		}

	  if (off > 0.0)
		{
          if ((road_index == 0) 
              && (segment_index == size_t (m_pit_in_index))
              && (segment->on_pit_merge (segment->length (), track_pos.y)))
            {
              road_index = 1;
              segment_index = 0;
            }
          else if ((road_index == 1) 
                   && (segment_index == mp_pit_lane->segments ().size () - 1))
            {
              road_index = 0;
              segment_index = m_pit_out_index;
            }
          else
            {
              segment_index++;
              if ((road_index == 0) && (segment_index == segments->size ()))
                segment_index = 0;
            }
        }
	  else
		{
          if ((road_index == 0) 
              && (segment_index == size_t (m_pit_out_index))
              && (segment->on_pit_merge (0.0, track_pos.y)))
            {
              road_index = 1;
              segment_index = mp_pit_lane->segments ().size () - 1;
            }
          else if ((road_index == 1) && (segment_index == 0))
            {
              road_index = 0;
              segment_index = m_pit_in_index;
            }
          else
            {
              if ((road_index == 0) && (segment_index == 0))
                segment_index = segments->size ();
              segment_index--;
			}
		}
      segments = &get_road (road_index).segments ();
	  segment = (*segments)[segment_index];
	  i++;
	}

  // Throw an exception if a segment could not be found.
  if (!found)
	{
      std::cerr << "not found" << std::endl;

	  throw Segment_Not_Found (world_pos);
	}

  assert (segment_index < segments->size ());
  m_material = segment->material_at (track_pos.x, track_pos.y);
  track_pos.x += segment->start_distance ();
  return track_pos;
}

const Road&
Strip_Track::get_road (size_t road_index) const
{
  switch (road_index)
    {
    case 0: return *mp_track;
    case 1: return *mp_pit_lane;
    default: assert (false);
    }
}

Contact_Info 
Strip_Track::test_for_contact (const Three_Vector& pos, 
							   double bump_parameter,
							   size_t& road_index,
							   size_t& segment_index)
{
  Three_Vector track_pos = track_coordinates (pos, road_index, segment_index);
  const Gl_Road_Segment* segment = 
    get_road (road_index).segments ()[segment_index];
  track_pos.x -= segment->start_distance ();

  bool contact = false;
  Three_Vector normal;

  // Test for contact with the road.
  double diff = 
    elevation (pos, bump_parameter, road_index, segment_index) - pos.z;
  if (diff >= 0.0)
	{
      //!! Let the bumps affect the normal x and y.
	  contact = true;
	  normal = segment->normal (track_pos.x, track_pos.y);
	}

  // Test for contact with the walls.
  // Left
  if (!contact)
	{
	  const Material& material = segment->left_material ();
	  double bump = material.bump (bump_parameter);
	  diff = track_pos.y - (segment->left_width (track_pos.x) + bump); 
	  if (diff >= 0.0)
		{
		  contact = true;
          m_material = material;
		  normal = segment->barrier_normal (track_pos.x, track_pos.y);
		}
	}

  // Right
  if (!contact)
	{
      const Material& material = segment->right_material ();
	  double bump = material.bump (bump_parameter);
	  diff = -track_pos.y - (segment->right_width (track_pos.x) + bump); 
	  if (diff >= 0.0)
		{
		  contact = true;
          m_material = material;
		  normal = segment->barrier_normal (track_pos.x, track_pos.y);
		}
	}

  return Contact_Info (contact, diff, normal, m_material);
}

// Return the distance along the track for the given position.
double 
Strip_Track::distance (const Three_Vector& pos, 
                       size_t& road_index,
                       size_t& segment_index)
{
  return track_coordinates (pos, road_index, segment_index).x;
}

Three_Vector
Strip_Track::position (double along, double from_center) const
{
  assert ((along >= 0.0) && (along <= mp_track->length ()));
  double distance = 0.0;
  for (Segment_List::const_iterator it = mp_track->segments ().begin ();
	   it != mp_track->segments ().end ();
	   it++)
	{
      if (distance + (*it)->length () >= along)
        {
          return (*it)->position (along - distance, from_center);
        }
      distance += (*it)->length ();
	}
  assert (false);
  return Three_Vector (0.0, 0.0, 0.0);
}

// Return the timing sector at the given distance.
int 
Strip_Track::sector (double distance)
{
  for (size_t i = 0; i < m_timing_lines.size (); i++)
	{
	  if (m_timing_lines [i] > distance)
		return i - 1;
	}
  return m_timing_lines.size () - 1;
}

void 
Strip_Track::set_start_direction (double degrees)
{
  m_start_direction = deg_to_rad (degrees);
  mp_track->set_start_direction (degrees);
}

void
Strip_Track::set_start_slope (double slope)
{
  mp_track->set_start_slope (slope);
}

void
Strip_Track::add_camera (const Camera& camera)
{
  m_cameras.push_back (camera);
}

double 
Strip_Track::camera_range (const Camera& camera) const
{
  double range = mp_track->segments ()[camera.segment_index]->start_distance () 
    + camera.position.x 
    - camera.range;
  return (range < 0.0) ? range + mp_track->length () : range; 
}

Three_Vector
Strip_Track::camera_position (const Camera& camera) const
{
  const Gl_Road_Segment& segment = *(mp_track->segments ()[camera.segment_index]);
  return segment.position (camera.position.x, camera.position.y)
    + Three_Vector (0.0, 0.0, camera.position.z);
}

const Camera& 
Strip_Track::get_camera (double distance) const
{
  if (m_cameras.empty ())
    return s_default_camera;
  
  // See if we're near the end of the track and should be picked up by
  // the first camera.
  std::vector <Camera>::const_iterator it = m_cameras.begin ();
  if (distance < camera_range (*it))
    {
      for (it = m_cameras.end () - 1; it > m_cameras.begin (); it--)
        {
          if (distance > camera_range (*it))
            break;
        }
    }
  return *it;
}

Three_Vector
Strip_Track::camera_target (const Camera& camera) const
{
  double angle = 
    mp_track->segments ()[camera.segment_index]->angle (camera.position.x);
  return camera_position (camera)
    + Three_Vector (-sin (deg_to_rad (camera.direction.x) + angle),
                    cos (deg_to_rad (camera.direction.x) + angle),
                    sin (deg_to_rad (camera.direction.y)));
}

const Rectangle&
Strip_Track::bounds () const
{
  return Rectangle (mp_track->bounds ()).enclose (mp_pit_lane->bounds ());
}
