//  Vamos Automotive Simulator
//  Copyright (C) 2001--2004 Sam Varner
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "../body/Car.h"
#include "../body/Fuel_Tank.h"
#include "../body/Wheel.h"
#include "../geometry/Conversions.h"
#include "../track/Track.h"
#include "Gl_World.h"
#include "Sounds.h"

#include <SDL/SDL.h>
#include <GL/glut.h>

#include <cmath>
#include <iomanip>
#include <sstream>
#include <string>
#include <cstdlib>
#include <algorithm>
#include <cctype>
#include <cassert>
#undef min

using namespace Vamos_Geometry;
using namespace Vamos_World;

//-----------------------------------------------------------------------------
static std::string 
format_time (double time)
{
  if (time == 0.0) return "";

  int minutes = int (time / 60.0);
  int seconds = int (time) % 60;
  int milliseconds = int (1000 * (time - floor (time)));

  std::ostringstream os;
  os << minutes << ':' << std::setfill ('0') << std::setw (2) 
     << seconds << '.' << milliseconds;
  return os.str ();
}

static std::string 
format_time_difference (double delta_time)
{
  if (delta_time == 0.0) return "";

  std::ostringstream os;
  if (delta_time > 0.0)
    {
      os << '+';
    }
  os << std::setprecision (3) << delta_time;
  return os.str ();
}

//-----------------------------------------------------------------------------
Gl_Window::Gl_Window (int width, int height, const char* name, bool full_screen)
  : m_video_flags (SDL_OPENGL | SDL_RESIZABLE | SDL_DOUBLEBUF)
{
  SDL_GL_SetAttribute (SDL_GL_STENCIL_SIZE, 1);
  if (full_screen)
    {
      m_video_flags |= SDL_FULLSCREEN;
      SDL_Rect** modes = SDL_ListModes (0, m_video_flags);
      if ((modes != 0) || (modes [0] != 0))
        {
          width = modes [0]->w;
          height = modes [0]->h;
        }
    }

  SDL_ShowCursor (false);
  SDL_WM_SetCaption (name, name);
  resize (width, height);
}


Gl_Window::~Gl_Window ()
{
}

void 
Gl_Window::resize (int width, int height)
{
  m_width = width;
  m_height = height;
  m_aspect = (m_height == 0) ? 1.0 : double (m_width) / m_height; 
  glViewport (0, 0, m_width, m_height);
  if (SDL_SetVideoMode (width, height, 0, m_video_flags) == 0)
    throw No_SDL_Screen (width, height, 0, m_video_flags);
}

//-----------------------------------------------------------------------------
Timer::Timer (double interval_to_average_over) :
  m_timeout (int (1000.0 * interval_to_average_over)),
  m_max_time_step (0.1),
  m_is_paused (false)
{
  start ();
}

void 
Timer::start ()
{
  m_start_time = SDL_GetTicks ();
  m_frames = 1;
}

void 
Timer::set_paused (bool is_paused)
{
  m_is_paused = is_paused;
}

void 
Timer::update ()
{
  if (m_is_paused)
    {
      m_start_time = SDL_GetTicks ();
      return;
    }

  int elapsed = SDL_GetTicks () - m_start_time;
  if (elapsed < m_timeout) return;
  m_time_step = std::min ((elapsed / 1000.0) / m_frames, 0.03);
  // Cap the time step at 0.03 second, 33 frames/s.  If it gets much longer
  // the simulation becomes unstable.
  start ();
}

//-----------------------------------------------------------------------------
Gl_World::Gl_World (int argc, char** argv,
                    Vamos_Track::Strip_Track* track, 
                    Atmosphere* atmosphere,
                    Sounds* sounds,
                    bool full_screen)
  : World (track, atmosphere),
    mp_sounds (sounds),
    mp_timer (0),
    mp_window (0),
    m_view (MAP_VIEW),
    mp_font (0)
{
  initialize_graphics (&argc, argv);
  mp_window = new Gl_Window (900, 600, argv [0], full_screen);
  m_mouse.set_axis_range (X, 0, mp_window->width ());
  m_mouse.set_axis_range (Y, 0, mp_window->height ());
  mp_timer = new Timer (0.1);
  set_attributes ();
  set_paused (true);
}

Gl_World::
~Gl_World ()
{
  delete mp_window;
  delete mp_timer;
  delete mp_sounds;
}

void 
Gl_World::initialize_graphics (int* argc, char** argv)
{
  if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_JOYSTICK) != 0) 
    throw Can_Not_Intialize_SDL (SDL_GetError ());
  atexit (SDL_Quit);
  SDL_JoystickOpen (0);
}

void 
Gl_World::set_attributes ()
{
  // Enable depth testing for hidden line removal
  glEnable (GL_DEPTH_TEST);
  // Allow transparency.
  glEnable (GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  // Allow textures.
  glEnable (GL_TEXTURE_2D);
  // Allow viewports.
  glEnable (GL_SCISSOR_TEST);
  // Allow masking.
  glEnable (GL_STENCIL_TEST);
  glClearStencil (0);
  // Allow lighting.
  glEnable (GL_LIGHTING);
  glEnable (GL_LIGHT0);
  // Create a Directional Light Source (x, y, z, w).
  GLfloat position [] = { 0.0, -1.0, 1.0, 0.0 };
  glLightfv (GL_LIGHT0, GL_POSITION, position);
  // Set the ambient light (R, G, B, A).
  GLfloat ambient [] = { 0.7, 0.7, 0.7, 1.0 };
  glLightfv (GL_LIGHT0, GL_AMBIENT, ambient);
  glClearColor (0.32, 0.65, 0.34, 0.0);
}

void 
Gl_World::add_car (Vamos_Body::Car* car)
{
  if (car == 0) return;
  World::add_car (car);
  m_view = BODY_VIEW;
  set_paused (false);
}

// Read the definition file.
void 
Gl_World::read (std::string data_dir, 
                std::string world_file,
                std::string controls_file)
{
  // Remember the file name for re-reading.
  if (data_dir != "")
    {
      m_data_dir = data_dir;
    }
  if (world_file != "")
    {
      m_world_file = world_file;
    }
  if (controls_file != "")
    {
      m_controls_file = controls_file;
    }

  World_Reader w_reader (m_world_file, this);
  Controls_Reader c_reader (m_controls_file, this);
}

bool 
Gl_World::fill_tank (double, double)
{
  focused_car ()->car->fuel_tank ()->fill ();
  return true;
}

void 
Gl_World::set_paused (bool is_paused)
{
  mp_timer->set_paused (is_paused);
  m_paused = is_paused;
}

bool 
Gl_World::pause (double, double)
{
  set_paused (!m_paused);
  return true;
}

bool 
Gl_World::quit (double, double)
{
  delete mp_window;
  mp_window = 0;
  std::exit (0);
  return true;
}

bool 
Gl_World::read_car (double, double)
{
  focused_car ()->car->read ();
  mp_sounds->set_engine_sound (focused_car ()->car->engine_sound ());
  focused_car ()->car->make_rear_view_mask (mp_window->width (), 
                                            mp_window->height ());
  return true;
}

bool 
Gl_World::read_track (double, double)
{
  mp_track->read ();
  display ();
  return true;
}

bool 
Gl_World::read_world (double, double)
{
  read ();
  return true;
}

bool 
Gl_World::reset_car (double, double)
{
  World::reset ();
  return true;
}

bool 
Gl_World::restart_car (double, double)
{
  World::restart ();
  return true;
}

bool 
Gl_World::start_engine (double, double)
{
  focused_car ()->car->start_engine ();
  return true;
}

bool 
Gl_World::cycle_view (double, double)
{
  switch (m_view)
    {
    case BODY_VIEW:
      m_view = MAP_VIEW;
      glClearColor (0.32, 0.65, 0.34, 0.0);
      break;
    case MAP_VIEW:
      m_view = WORLD_VIEW;
      break;
    case WORLD_VIEW:
    default:
      m_view = BODY_VIEW;
      break;
    }

  return true;
}

bool 
Gl_World::initial_shift_up (double, double)
{
  if (focused_car ()->car->gear () == 0)
    {
      focused_car ()->car->shift_up ();
      return true;
    }
  return false;
}

bool 
Gl_World::initial_shift_down (double, double)
{
  if (focused_car ()->car->gear () == 0)
    {
      focused_car ()->car->shift_down ();
      return true;
    }
  return false;
}

bool 
Gl_World::shift_up (double, double)
{
  if (focused_car ()->car->gear () != 0)
    {
      focused_car ()->car->shift_up ();
      return true;
    }
  return false;
}

bool 
Gl_World::shift_down (double, double)
{
  if (focused_car ()->car->gear () != 0)
    {
      focused_car ()->car->shift_down ();
      return true;
    }
  return false;
}

bool 
Gl_World::initial_shift_up_disengage (double, double time)
{
  if (focused_car ()->car->gear () == 0)
    {
      focused_car ()->car->disengage_clutch (time);
      focused_car ()->car->shift_up ();
      return true;
    }
  return false;
}

bool 
Gl_World::initial_shift_down_disengage (double, double time)
{
  if (focused_car ()->car->gear () == 0)
    {
      focused_car ()->car->disengage_clutch (time);
      focused_car ()->car->shift_down ();
      return true;
    }
  return false;
}

bool 
Gl_World::shift_up_disengage (double, double time)
{
  if (focused_car ()->car->gear () != 0)
    {
      if (!focused_car ()->car->fast_shift ())
        {
          focused_car ()->car->disengage_clutch (time);
        }
      focused_car ()->car->shift_up ();
      return true;
    }
  return false;
}

bool 
Gl_World::shift_down_disengage (double, double time)
{
  if (focused_car ()->car->gear () != 0)
    {
      if (!focused_car ()->car->fast_shift ())
        {
          focused_car ()->car->disengage_clutch (time);
        }
      focused_car ()->car->shift_down ();
      return true;
    }
  return false;
}

bool 
Gl_World::initial_clutch (double value, double)
{
  if (focused_car ()->car->last_gear () == 0)
    {
      focused_car ()->car->clutch (value);
      return true;
    }
  return false;
}

bool 
Gl_World::clutch (double value, double)
{
  if (focused_car ()->car->last_gear () != 0)
    {
      focused_car ()->car->clutch (value);
      return true;
    }
  return false;
}

bool 
Gl_World::initial_engage_clutch (double, double time)
{
  if (focused_car ()->car->last_gear () == 0)
    {
      focused_car ()->car->engage_clutch (time);
      return true;
    }
  return false;
}

bool 
Gl_World::initial_disengage_clutch (double, double time)
{
  if (focused_car ()->car->last_gear () == 0)
    {
      focused_car ()->car->disengage_clutch (time);
      return true;
    }
  return false;
}

bool 
Gl_World::engage_clutch (double, double time)
{
  if (focused_car ()->car->last_gear () != 0)
    {
      focused_car ()->car->engage_clutch (time);
      return true;
    }
  return false;
}

bool 
Gl_World::disengage_clutch (double, double time)
{
  if (focused_car ()->car->last_gear () != 0)
    {
      focused_car ()->car->disengage_clutch (time);
      return true;
    }
  return false;
}

bool 
Gl_World::gas (double value, double time)
{
  focused_car ()->car->gas (value, time);
  return false;
}

bool 
Gl_World::brake (double value, double time)
{
  focused_car ()->car->brake (value, time);
  return false;
}

bool 
Gl_World::steer (double value, double time)
{
  focused_car ()->car->steer (value, time);
  return false;
}

bool 
Gl_World::steer_left (double value, double time)
{
  focused_car ()->car->steer (value, time);
  return true;
}

bool 
Gl_World::steer_right (double value, double time)
{
  focused_car ()->car->steer (-value, time);
  return true;
}

bool 
Gl_World::pan_left (double value, double time)
{
  focused_car ()->car->pan (value, time);
  return true;
}

bool 
Gl_World::pan_right (double value, double time)
{
  focused_car ()->car->pan (-value, time);
  return true;
}

bool 
Gl_World::focus_previous_car (double, double)
{
  focus_other_car (-1);
  return true;
}

bool 
Gl_World::focus_next_car (double, double)
{
  focus_other_car (1);
  return true;
}

// The main loop of the simulation
void 
Gl_World::animate ()
{
  if (focused_car () != 0)
    {
      for (int loop = 0; loop < mp_timer->steps_per_frame (); loop++)
        {
          propagate_cars (mp_timer->get_substep ());
        }
      play_sounds ();
      update_car_timing ();
    }
  mp_timer->add_frame ();
}

void
Gl_World::propagate_cars (double time_step)
{
  for (std::vector <Car_Information>::iterator it = m_cars.begin ();
       it != m_cars.end ();
       it++)
    {
      it->car->propagate (time_step);
      interact (it->car, it->road_index, it->segment_index);

      for (std::vector <Car_Information>::iterator other = m_cars.begin ();
           other != m_cars.end ();
           other++)
          {
            if (other != it)
                collide (it->car, other->car);
          }
    }
}

void
Gl_World::update_car_timing ()
{
  for (std::vector <Car_Information>::iterator it = m_cars.begin ();
       it != m_cars.end ();
       it++)
    {
      const double distance = 
        mp_track->distance (it->car->chassis ().position (), 
                            it->road_index,
                            it->segment_index);
      const int sector = mp_track->sector (distance);
      it->timing.update (mp_timer->get_time_step (), distance, sector);
    }
}

void
Gl_World::play_sounds ()
{
  if (mp_sounds == 0) return;

  double tire_slide = 0.0;
  double grass_speed = 0.0;
  double gravel_speed = 0.0;
  double scrape_speed = 0.0;
  double hard_crash_speed = 0.0;
  double soft_crash_speed = 0.0;
  
  for (std::vector <Interaction_Info>::const_iterator 
         it = m_interaction_info.begin ();
       it != m_interaction_info.end ();
       it++)
    {
      if (it->car != focused_car ()->car) continue;

      switch (it->track_material)
        {
        case Material::ASPHALT:
        case Material::CONCRETE:
        case Material::METAL:
          if (it->car_material == Material::RUBBER)
            {
              tire_slide = focused_car ()->car->slide ();
            }
          else if (it->car_material == Material::METAL)
            {
              scrape_speed = it->parallel_speed;
              hard_crash_speed = it->perpendicular_speed;
            }
          break;
        case Material::GRASS:
          grass_speed = it->parallel_speed;
          break;
        case Material::GRAVEL:
          gravel_speed = it->parallel_speed;
          break;
        case Material::RUBBER:
          soft_crash_speed = it->perpendicular_speed;
          break;
        default:
          break;
        }
    }
  m_interaction_info.clear ();

  mp_sounds->play_engine_sound (focused_car ()->car->engine_pitch (),
                                focused_car ()->car->engine_volume (),
                                focused_car ()->car->chassis ().
                                position (focused_car ()->car->engine ()),
                                focused_car ()->car->chassis ().
                                velocity (focused_car ()->car->engine ()));

  mp_sounds->play_tire_squeal_sound (tire_slide,
                                     focused_car ()->car->chassis ().position ());
  mp_sounds->play_grass_sound (grass_speed,
                               focused_car ()->car->chassis ().position ());
  mp_sounds->play_gravel_sound (gravel_speed,
                                focused_car ()->car->chassis ().position ());
  mp_sounds->play_scrape_sound (scrape_speed,
                                focused_car ()->car->chassis ().position ());
  mp_sounds->play_wind_sound ((focused_car ()->car->chassis ().
                               cm_velocity () 
                               - mp_atmosphere->velocity ()).magnitude (),
                              focused_car ()->car->chassis ().position ());
  mp_sounds->play_hard_crash_sound (hard_crash_speed,
                                    focused_car ()->car->chassis ().position ());
  mp_sounds->play_soft_crash_sound (soft_crash_speed,
                                    focused_car ()->car->chassis ().position ());
}

void 
Gl_World::show_full_window ()
{
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();

  glViewport (0, 0, mp_window->width (), mp_window->height ());
  glScissor (0, 0, mp_window->width (), mp_window->height ());
  glStencilFunc (GL_ALWAYS, 1, 1);
  glStencilOp (GL_KEEP, GL_KEEP, GL_KEEP);

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void 
Gl_World::draw_mirror_views ()
{
  for (int i = 0; i < focused_car ()->car->get_n_mirrors (); i++)
    {
      focused_car ()->car->draw_rear_view (mp_window->aspect (), i);

       glMatrixMode (GL_MODELVIEW);

       // Enable the rearview mirror mask.
       glStencilFunc (GL_EQUAL, 1, 1);

      // Front and back are reversed in the mirrors, so cull front
      // faces while we're drawing mirror views.
      glPushAttrib (GL_POLYGON_BIT);
      {
        glCullFace (GL_FRONT);
        mp_track->draw_sky (focused_car ()->car->view_position ());
        mp_track->draw ();
        draw_cars (false, false);
      }
      glPopAttrib ();
    }
}

void 
Gl_World::set_car_view (Vamos_Body::Car* car)
{
  assert (car != 0);
  car->set_perspective (mp_window->aspect ());
  car->view ();
}

void 
Gl_World::set_map_view ()
{
  // Adjust the mins and maxes to keep the correct aspect ratio of the
  // track regardless of the window's size.
  Rectangle bounds (mp_track->bounds ());
  double ratio = bounds.aspect () / mp_window->aspect ();
  if (ratio < 1.0)
    {
      // The window is wider than the track.  Stretch the x-dimension.
      bounds.left /= ratio;
      bounds.right /= ratio;
    }
  else
    {
      // The track is wider than the window.  Stretch the y-dimension.
      bounds.top *= ratio;
      bounds.bottom *= ratio;
    }
  glOrtho (bounds.left, bounds.right, bounds.bottom, bounds.top, -1000, 1000);
}

void 
Gl_World::set_world_view (const Vamos_Track::Camera& camera)
{
  gluPerspective (camera.vertical_field_angle, 
                  mp_window->aspect (), 
                  10.0, 1000.0);
  Three_Vector camera_position = mp_track->camera_position (camera);
  Three_Vector target_position = camera.fixed
    ? mp_track->camera_target (camera)
    : focused_car ()->car->chassis ().cm_position ();
  gluLookAt (camera_position.x, camera_position.y, camera_position.z,
             target_position.x, target_position.y, target_position.z,
             0.0, 0.0, 1.0); // up direction

  Three_Vector direction (target_position - camera_position);
  float at_up [6] = { direction.x, direction.y, direction.z, 0.0, 0.0, 1.0 };

  alListener3f (AL_POSITION, 
                camera_position.x, camera_position.y, camera_position.z);
  alListener3f (AL_VELOCITY, 0.0, 0.0, 0.0);
  alListenerfv (AL_ORIENTATION, at_up);
}

void 
Gl_World::draw_track (bool draw_sky, const Three_Vector& view_position)
{
  glMatrixMode (GL_MODELVIEW);
  if (draw_sky)
    {
      assert (focused_car () != 0);
      mp_track->draw_sky (view_position);
    }
  else
    {
      mp_track->draw_map_background ();
    }
  mp_track->draw ();
}

void 
Gl_World::draw_cars (bool draw_interior, bool draw_focused_car)
{
  for (std::vector <Car_Information>::iterator it = m_cars.begin ();
       it != m_cars.end ();
       it++)
    {
      assert (it->car != 0);
      if (it->car != focused_car ()->car)
        {
          it->car->draw ();
          if (draw_interior)
            {
              it->car->draw_interior ();
            }
        }
    }
  if (draw_focused_car)
    {
      focused_car ()->car->draw ();
      if (draw_interior)
        {
          focused_car ()->car->draw_interior ();
        }
    }
}

void 
Gl_World::show_scene ()
{
  glFlush ();
  SDL_GL_SwapBuffers ();
}

void 
Gl_World::display ()
{
  if (m_view == BODY_VIEW)
    focused_car ()->car->update_rear_view_mask (mp_window->width (), 
                                                mp_window->height ());
  show_full_window ();

  switch (m_view)
    {
    case BODY_VIEW:
      set_car_view (focused_car ()->car);
      draw_track (true, focused_car ()->car->view_position ());
      draw_cars (true);
      draw_timing_info ();
      draw_mirror_views ();
      break;
    case MAP_VIEW:
      set_map_view ();
      draw_track (false);
      if (focused_car () != 0)
        {
          draw_cars (false);
        }
      break;
    case WORLD_VIEW:
      {
        const Vamos_Track::Camera& camera =
          mp_track->get_camera (focused_car ()->timing.get_distance ());
        set_world_view (camera);
        draw_track (true, mp_track->camera_position (camera));
        draw_cars (true);
      }
      break;
    }

  show_scene ();
}

void 
Gl_World::reshape (int width, int height)
{
  mp_window->resize (width, height);
#warning "duplicate set_axis_range ()"
  m_mouse.set_axis_range (X, 0, width);
  m_mouse.set_axis_range (Y, 0, height);
  if (focused_car () != 0)
    {
      focused_car ()->car->make_rear_view_mask (width, height);
    }
}

void 
Gl_World::draw_string (const std::string& str, double x, double y)
{
  glRasterPos2d (x, y);
  for (std::string::const_iterator it = str.begin (); it != str.end (); it++)
    {
      glutBitmapCharacter (GLUT_BITMAP_8_BY_13, *it);
    }
}

void 
Gl_World::draw_timing_info ()
{
  if (focused_car ()->car == 0) return;

  glDisable (GL_DEPTH_TEST);
  glDisable (GL_LIGHTING);
  glDisable (GL_TEXTURE_2D);

  glColor3f (1.0, 1.0, 1.0);
  
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  
  std::ostringstream b_stream;
  double x = 5.5;

  gluOrtho2D (0, 10, 0, 10);

  const Timing_Info& timing = focused_car ()->timing;
  b_stream << "Lap Time " << format_time (timing.get_lap_time ());
  draw_string (b_stream.str (), x, 1.4);

  b_stream.str ("");
  b_stream << "    Last " << format_time (timing.get_previous_lap_time ()) 
           << " " 
           << format_time_difference (timing.get_lap_time_difference ());
  draw_string (b_stream.str (), x, 1.0);

  b_stream.str ("");
  b_stream << "    Best " << format_time (timing.get_best_lap_time ());
  draw_string (b_stream.str (), x, 0.6);

  b_stream.str ("");
  b_stream << 1.0 / mp_timer->get_time_step () << " frame/s";
  draw_string (b_stream.str (), x, 0.2);

  x = 7.5;

  int sector = timing.get_sector ();
  b_stream.str ("");
  b_stream << "   Sector ";
  if (sector != -1)
    {
      b_stream << sector + 1 << " " 
               << format_time (timing.get_sector_time ());
    }
  draw_string (b_stream.str (), x, 1.4);

  b_stream.str ("");
  b_stream << "       Best "; 
  if (sector != -1)
    {
      b_stream << format_time (timing.get_best_sector_time (sector));
    }
  draw_string (b_stream.str (), x, 1.0);

  b_stream.str ("");
  b_stream << "Last Sector ";
  if (timing.get_previous_sector () != -1)
    {
      b_stream << format_time (timing.get_previous_sector_time ()) << "  " 
               << format_time_difference 
        (timing.get_previous_sector_time_difference ());
    }
  draw_string (b_stream.str (), x, 0.6);

  b_stream.str ("");
  b_stream << "Distance " << int (timing.get_distance ()) << " m";
  draw_string (b_stream.str (), x, 0.2);

  glEnable (GL_DEPTH_TEST);
  glEnable (GL_LIGHTING);
  glEnable (GL_TEXTURE_2D);
}

void 
Gl_World::start ()
{
  SDL_Event event;
  while (true)
    {
      mp_timer->update ();
      while (SDL_PollEvent (&event))
        {
          switch (event.type) 
            {
            case SDL_JOYAXISMOTION:
              m_joystick.move (event.jaxis.axis, event.jaxis.value);
              break;
            case SDL_JOYBUTTONDOWN:
              m_joystick.press (event.jbutton.button + 1);
              break;
            case SDL_JOYBUTTONUP:
              m_joystick.release (event.jbutton.button + 1);
              break;
            case SDL_KEYDOWN:
              m_keyboard.press (event.key.keysym.sym);
              break;                
            case SDL_KEYUP:
              m_keyboard.release (event.key.keysym.sym);
              break;
            case SDL_MOUSEMOTION:
              m_mouse.move (X, event.motion.x);
              m_mouse.move (Y, event.motion.y);
              break;
            case SDL_MOUSEBUTTONDOWN:
              m_mouse.press (event.button.button);
              break;
            case SDL_MOUSEBUTTONUP:
              m_mouse.release (event.button.button);
              break;
            case SDL_VIDEORESIZE:
              reshape (event.resize.w, event.resize.h);
              break;
            case SDL_QUIT:
              quit ();
              break;
           }
        }

      if (m_paused)
        {
          mp_sounds->pause ();
          SDL_Delay (100);
        }
      else
        {
          SDL_Delay (0);
          animate ();
        }
      display ();
    }
}

void 
Gl_World::set_focused_car (size_t index)
{
  World::set_focused_car (index);

  if (focused_car () != 0)
    {
      mp_sounds->set_engine_sound (focused_car ()->car->engine_sound ());
      focused_car ()->car->make_rear_view_mask (mp_window->width (), 
                                                mp_window->height ());
    }
}

//* Class World_Reader

//** Constructor

World_Reader::World_Reader (std::string file_name, Gl_World* world) 
  : mp_world (world)
{
  read (file_name);
}

void 
World_Reader::on_start_tag (const Vamos_Geometry::XML_Tag& tag)
{
}

void 
World_Reader::on_end_tag (const Vamos_Geometry::XML_Tag& tag)
{
}

void 
World_Reader::on_data (std::string data)
{
  if (data.size () == 0)
    {
      return;
    }
  std::istringstream is (data.c_str ());

  if (path () == "/world/gravity")
    {
      double grav;
      is >> grav;
      mp_world->gravity (grav);
    }
  else if (path () == "/world/maximum-time-step")
    {
      double step;
      is >> step;
      mp_world->set_max_time_step (step);
    }
  else if (path () == "/world/atmosphere/density")
    {
      double density;
      is >> density;
      mp_world->mp_atmosphere->density (density);
    }
  else if (path () == "/world/atmosphere/velocity")
    {
      Three_Vector velocity;
      is >> velocity;
      mp_world->mp_atmosphere->velocity (velocity);
    }
  else if (path () == "/world/atmosphere/speed-of-sound")
    {
      float v_s;
      is >> v_s;
      alSpeedOfSound (v_s);
    }
  else if (path () == "/world/lighting/source-position")
    {
      Three_Vector pos;
      is >> pos;
      GLfloat position [] = { pos.x, pos.y, pos.z, 0.0 };
      glLightfv (GL_LIGHT0, GL_POSITION, position);
    }
  else if (path () == "/world/lighting/ambient")
    {
      Three_Vector amb;
      is >> amb;
      GLfloat ambient [] = { amb.x, amb.y, amb.z, 1.0 };
      glLightfv (GL_LIGHT0, GL_AMBIENT, ambient);
    }
}

//* Class Controls_Reader
static int 
translate_key (std::string key_name)
{
  // If key_name is a single character, return its integer value...
  if (key_name.size () == 1)
    return int (key_name [0]);

  // ...otherwise, try to interpret it as the name of a special key.

  // Downcase the string to do case-insensinive comparisons.
  std::transform (key_name.begin (), key_name.end (), key_name.begin (),
                  ::tolower);

  if (key_name == "escape")
      return 27;
  if (key_name == "delete")
      return 127;
  if (key_name == "up")
      return SDLK_UP;
  if (key_name == "down")
      return SDLK_DOWN;
  if (key_name == "left")
      return SDLK_LEFT;
  if (key_name == "right")
      return SDLK_RIGHT;
  if (key_name == "insert")
      return SDLK_INSERT;
  if (key_name == "home")
      return SDLK_HOME;
  if (key_name == "end")
      return SDLK_END;
  if (key_name == "page up")
      return SDLK_PAGEUP;
  if (key_name == "page down")
      return SDLK_PAGEDOWN;
  if (key_name == "f1")
      return SDLK_F1;
  if (key_name == "f2")
      return SDLK_F2;
  if (key_name == "f3")
      return SDLK_F3;
  if (key_name == "f4")
      return SDLK_F4;
  if (key_name == "f5")
      return SDLK_F5;
  if (key_name == "f6")
      return SDLK_F6;
  if (key_name == "f7")
      return SDLK_F7;
  if (key_name == "f8")
      return SDLK_F8;
  if (key_name == "f9")
      return SDLK_F9;
  if (key_name == "f10")
      return SDLK_F10;
  if (key_name == "f11")
      return SDLK_F11;
  if (key_name == "f12")
      return SDLK_F12;

  assert (false);
  return 0;
}

static int 
mouse_button_to_control (std::string button)
{
  if ((button == "middle") || (button == "Middle") || (button == "MIDDLE"))
      return SDL_BUTTON_MIDDLE;
  if ((button == "right") || (button == "Right") || (button == "RIGHT"))
      return SDL_BUTTON_RIGHT;
  return SDL_BUTTON_LEFT;
}

//** Constructor

Controls_Reader::Controls_Reader (std::string file_name, Gl_World* world) 
  : mp_world (world)
{
  SDL_ShowCursor (false);
  // Turn on if we bind mouse motion events.

  m_function_map ["pause"] = (Callback_Function)&Gl_World::pause;
  m_function_map ["quit"] = (Callback_Function)&Gl_World::quit;
  m_function_map ["read track"] = (Callback_Function)&Gl_World::read_track;
  m_function_map ["read world"] = (Callback_Function)&Gl_World::read_world;
  m_function_map ["cycle view"] = (Callback_Function)&Gl_World::cycle_view;

  m_function_map ["reset car"] = (Callback_Function)&Gl_World::reset_car;
  m_function_map ["restart car"] = (Callback_Function)&Gl_World::restart_car;
  m_function_map ["read car"] = (Callback_Function)&Gl_World::read_car;
  m_function_map ["start engine"] = (Callback_Function)&Gl_World::start_engine;
  m_function_map ["fill tank"] = (Callback_Function)&Gl_World::fill_tank;

  m_function_map ["initial shift up"] = (Callback_Function)&Gl_World::initial_shift_up;
  m_function_map ["initial shift down"] = (Callback_Function)&Gl_World::initial_shift_down;
  m_function_map ["shift up"] = (Callback_Function)&Gl_World::shift_up;
  m_function_map ["shift down"] = (Callback_Function)&Gl_World::shift_down;
  m_function_map ["initial shift up disengage"] 
    = (Callback_Function)&Gl_World::initial_shift_up_disengage;
  m_function_map ["initial shift down disengage"] 
    = (Callback_Function)&Gl_World::initial_shift_down_disengage;
  m_function_map ["shift up disengage"] 
    = (Callback_Function)&Gl_World::shift_up_disengage;
  m_function_map ["shift down disengage"] 
    = (Callback_Function)&Gl_World::shift_down_disengage;
  m_function_map ["initial engage clutch"] 
    = (Callback_Function)&Gl_World::initial_engage_clutch;
  m_function_map ["initial disengage clutch"] 
    = (Callback_Function)&Gl_World::initial_disengage_clutch;
  m_function_map ["engage clutch"] = (Callback_Function)&Gl_World::engage_clutch;
  m_function_map ["disengage clutch"] = (Callback_Function)&Gl_World::disengage_clutch;
  m_function_map ["initial clutch"] = (Callback_Function)&Gl_World::initial_clutch;
  m_function_map ["clutch"] = (Callback_Function)&Gl_World::clutch;

  m_function_map ["steer"] = (Callback_Function)&Gl_World::steer;
  m_function_map ["steer right"] = (Callback_Function)&Gl_World::steer_right;
  m_function_map ["steer left"] = (Callback_Function)&Gl_World::steer_left;
  m_function_map ["gas"] = (Callback_Function)&Gl_World::gas;
  m_function_map ["brake"] = (Callback_Function)&Gl_World::brake;

  m_function_map ["pan left"] = (Callback_Function)&Gl_World::pan_left;
  m_function_map ["pan right"] = (Callback_Function)&Gl_World::pan_right;

  m_function_map ["focus previous car"] = (Callback_Function)&Gl_World::focus_previous_car;
  m_function_map ["focus next car"] = (Callback_Function)&Gl_World::focus_next_car;

  read (file_name);
}

void 
Controls_Reader::on_start_tag (const Vamos_Geometry::XML_Tag& tag)
{
  if (label () == "bind")
    {
      m_function = "";
      m_control = 0;
      m_direction = NONE;
      m_factor = 1.0;
      m_offset = 0.0;
      m_deadband = 0.0;
      m_upper_deadband = 0.0;
      m_time = 0.0;
    }
}

void 
Controls_Reader::on_end_tag (const Vamos_Geometry::XML_Tag& tag)
{
  if (label () == "up")
    m_direction = UP;
  else if (label () == "down")
    m_direction = DOWN;
  else if (label () == "left")
    m_direction = LEFT;
  else if (label () == "right")
    m_direction = RIGHT;
  else if (label () == "forward")
    m_direction = FORWARD;
  else if (label () == "backward")
    m_direction = BACKWARD;

  else if (label () == "bind")
    {
      std::map <std::string, Callback_Function>
        ::iterator it = m_function_map.find (m_function);

      if (it == m_function_map.end ())
        {
          throw Unknown_Function (m_function);
        }
      // Register the callback.
      switch (m_type)
        {
        case KEY:
          mp_world->m_keyboard.
            bind_action (m_control,
                         m_direction,
                         (Control_Handler*)mp_world,
                         it->second,
                         m_time);
          break;
        case JOYSTICK_BUTTON:
          mp_world->m_joystick.
            bind_action (m_control,
                         m_direction,
                         (Control_Handler*)mp_world,
                         it->second, 
                         m_time);
          break;
        case JOYSTICK_AXIS:
          mp_world->m_joystick.bind_motion (m_control, 
                                            m_direction,
                                            (Control_Handler*)mp_world,
                                            it->second,
                                            m_factor, 
                                            m_offset, 
                                            m_deadband,
                                            m_upper_deadband);
          break;
        case MOUSE_BUTTON:
          mp_world->m_mouse.
            bind_action (m_control, 
                         m_direction,
                         (Control_Handler*)mp_world,
                         it->second, 
                         m_time);
          break;
        case MOUSE_MOTION:
          SDL_ShowCursor (true);
          mp_world->m_mouse.bind_motion (m_control,
                                         m_direction,
                                         (Control_Handler*)mp_world,
                                         it->second,
                                         m_factor, 
                                         m_offset, 
                                         m_deadband,
                                         m_upper_deadband);
          break;
        default:
          assert (false);
        }
    }
}

void 
Controls_Reader::on_data (std::string data)
{
  if (data.size () == 0) return;

  std::istringstream is (data.c_str ());
  
  if (label () == "key")
    {
      m_type = KEY;
      std::string key;
      is >> key;
      m_control = translate_key (key);
    }
  else if (label () == "button")
    {
      m_type = JOYSTICK_BUTTON;
      is >> m_control;
    }
  else if (label () == "mouse-button")
    {
      m_type = MOUSE_BUTTON;
      std::string button;
      is >> button;
      m_control = mouse_button_to_control (button);
    }
  else if (label () == "axis")
    {
      m_type = JOYSTICK_AXIS;
      std::string axis;
      is >> m_control;
    }
  else if (label () == "mouse-direction")
    {
      m_type = MOUSE_MOTION;
      std::string axis;
      is >> m_control;
    }
  else if (label () == "function")
      m_function = data;
  else if (label () == "factor")
      is >> m_factor;
  else if (label () == "offset")
      is >> m_offset;
  else if (label () == "deadband")
      is >> m_deadband;
  else if (label () == "upper-deadband")
      is >> m_upper_deadband;
  else if (label () == "time")
      is >> m_time;
}
